#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#define MEMORY_REGION_SIZE 0x10000
#define PROG_OFFSET     0  * MEMORY_REGION_SIZE
#define RAM_OFFSET      1  * MEMORY_REGION_SIZE
#define USERDATA_OFFSET 15 * MEMORY_REGION_SIZE
#define END_OFFSET      16 * MEMORY_REGION_SIZE

#define NUM_REGS 16
#define REG_IP 15

#define OPCODE_SET_FLAGS_MASK (1 << 28)
#define OPCODE_OPCODE_MASK 0x0FFFFFFF

#define COND_MODE_OFFSET 28
#define COND_MODE_MASK   (0xF << COND_MODE_OFFSET)
#define COND_FLAG_MASK   (~COND_MODE_MASK)

#define VALIDATE_REG_NUM(x) { if (x >= NUM_REGS) { return 1; } }
#define VALIDATE_ADDRESS(x, vm) { if (x >= vm->memory_size - sizeof(uint32_t)) { return 1; } }

enum conditional_modes {
  COND_ALWAYS = 0,
  COND_AND,
  COND_OR,
};

enum flags {
  EQ_ZERO = 0,
  LT_NEG,
  GT_POS,
  FLAGS_END,
};

enum instructions {
  OP_MOV,

  OP_ADD,
  OP_SUB,

  OP_MUL,
  OP_DIV,
  OP_MOD,

  OP_ORR,
  OP_AND,
  OP_XOR,

  OP_NOT,
  OP_MHI,

  OP_MVI,

  OP_LMW,
  OP_SMW,

  OP_LMB,
  OP_SMB,

  OP_EXT,
};

typedef struct _vm_state {
  uint32_t regs[NUM_REGS];
  uint32_t flags;
  uint8_t *memory;
  size_t memory_size;
} vm_state_t;

typedef struct _instruction {
  uint32_t opcode;
  uint32_t conditional;
  uint32_t arg1;
  uint32_t arg2;
} instruction_t;

uint32_t calculate_flags(uint32_t result) {
  uint32_t flags = 0;
  flags |= (result == 0) << EQ_ZERO;
  flags |= (*(int32_t *)&result <  0) << LT_NEG;
  flags |= (*(int32_t *)&result >  0) << GT_POS;
  return flags;
}

int test_conditionals(uint8_t cond_mode, uint32_t cond_flags, uint32_t flags) {
  uint8_t should_execute;
  uint8_t flag;

  switch (cond_mode) {
  case COND_ALWAYS:
    should_execute = 1;
    break;
  case COND_AND:
    should_execute = 1;
    for (flag = 0; flag < FLAGS_END; flag++) {
      if (cond_flags & (1 << (2 * flag))) {
        if (!(flags & (1 << flag))) {
          should_execute = 0;
          break;
        }
      }

      if (cond_flags & (1 << ((2 * flag) + 1))) {
        if (flags & (1 << flag)) {
          should_execute = 0;
          break;
        }
      }
    }
    break;
  case COND_OR:
    should_execute = 0;
    for (flag = 0; flag < FLAGS_END; flag++) {
      if (cond_flags & (1 << (2 * flag))) {
        if (flags & (1 << flag)) {
          should_execute = 1;
          break;
        }
      }

      if (cond_flags & (1 << ((2 * flag) + 1))) {
        if (!(flags & (1 << flag))) {
          should_execute = 1;
          break;
        }
      }
    }
    break;
  default:
    return -1;
  }

  return should_execute;
}

int run_vm(vm_state_t *vm) {
  uint32_t ip;
  instruction_t instr;

  uint8_t cond_mode;
  uint32_t cond_flags;
  int should_execute;

  uint32_t op_result;
  uint8_t should_set_flags;

  uint64_t intermediate;

  while (42) {
    ip = vm->regs[REG_IP];
    if (ip > (vm->memory_size - sizeof(instr))) {
      return 1;
    }
    memcpy(&instr, &vm->memory[ip], sizeof(instr));

    cond_mode  = (instr.conditional & COND_MODE_MASK) >> COND_MODE_OFFSET;
    cond_flags = (instr.conditional & COND_FLAG_MASK);

    should_execute = test_conditionals(cond_mode, cond_flags, vm->flags);
    if (should_execute < 0) {
      return 1;
    }

    op_result = 0;
    should_set_flags = (instr.opcode & OPCODE_SET_FLAGS_MASK) != 0;

    if (should_execute) {
      switch (instr.opcode & OPCODE_OPCODE_MASK) {
      case OP_MOV:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        vm->regs[instr.arg1]  = vm->regs[instr.arg2];
        op_result = vm->regs[instr.arg1];
        break;
      case OP_ADD:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        vm->regs[instr.arg1] += vm->regs[instr.arg2];
        op_result = vm->regs[instr.arg1];
        break;
      case OP_SUB:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        vm->regs[instr.arg1] -= vm->regs[instr.arg2];
        op_result = vm->regs[instr.arg1];
        break;
      case OP_MUL:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        vm->regs[instr.arg1] *= vm->regs[instr.arg2];
        op_result = vm->regs[instr.arg1];
        break;
      case OP_DIV:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        vm->regs[instr.arg1] /= vm->regs[instr.arg2];
        op_result = vm->regs[instr.arg1];
        break;
      case OP_MOD:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        vm->regs[instr.arg1] %= vm->regs[instr.arg2];
        op_result = vm->regs[instr.arg1];
        break;
      case OP_ORR:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        vm->regs[instr.arg1] |= vm->regs[instr.arg2];
        op_result = vm->regs[instr.arg1];
        break;
      case OP_AND:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        vm->regs[instr.arg1] &= vm->regs[instr.arg2];
        op_result = vm->regs[instr.arg1];
        break;
      case OP_XOR:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        vm->regs[instr.arg1] ^= vm->regs[instr.arg2];
        op_result = vm->regs[instr.arg1];
        break;

      case OP_NOT:
        VALIDATE_REG_NUM(instr.arg1);
        vm->regs[instr.arg1] = ~vm->regs[instr.arg1];
        op_result = vm->regs[instr.arg1];
        break;
      case OP_MHI:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        intermediate = vm->regs[instr.arg1] * vm->regs[instr.arg2];
        vm->regs[instr.arg1] = intermediate >> 32;
        op_result = vm->regs[instr.arg1];
        break;

      case OP_MVI:
        VALIDATE_REG_NUM(instr.arg1);
        vm->regs[instr.arg1] = instr.arg2;
        op_result = vm->regs[instr.arg1];
        break;

      case OP_LMW:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        VALIDATE_ADDRESS(vm->regs[instr.arg2], vm);
        vm->regs[instr.arg1] = *(uint32_t *)&vm->memory[vm->regs[instr.arg2]];
        op_result = vm->regs[instr.arg1];
        break;
      case OP_SMW:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        VALIDATE_ADDRESS(vm->regs[instr.arg1], vm);
        *(uint32_t *)&vm->memory[vm->regs[instr.arg1]] = vm->regs[instr.arg2];
        should_set_flags = 0;
        break;

      case OP_LMB:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        VALIDATE_ADDRESS(vm->regs[instr.arg2], vm);
        vm->regs[instr.arg1] = vm->memory[vm->regs[instr.arg2]];
        op_result = vm->regs[instr.arg1];
        break;
      case OP_SMB:
        VALIDATE_REG_NUM(instr.arg1);
        VALIDATE_REG_NUM(instr.arg2);
        VALIDATE_ADDRESS(vm->regs[instr.arg1], vm);
        vm->memory[vm->regs[instr.arg1]] = (uint8_t)vm->regs[instr.arg2];
        should_set_flags = 0;
        break;

      case OP_EXT:
        return 0;
      default:
        return 1;
      }
    }

    if (should_set_flags) {
      vm->flags = calculate_flags(op_result);
    }

    vm->regs[REG_IP] += sizeof(instr);
  }
  return 0;
}

int main(int argc, char **argv) {
  FILE *f;
  vm_state_t vm;
  long file_length;
  uint32_t i;

  if (argc != 3) {
    fprintf(stderr, "Usage: %s <prog file> <user data>\n", argv[0]);
    return 1;
  }

  if ((f = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "Unable to open program.\n");
    return 2;
  }

  fseek(f, 0, SEEK_END);
  file_length = ftell(f);
  if (file_length > MEMORY_REGION_SIZE) {
    fprintf(stderr, "Program too large.\n");
    return 3;
  }
  fseek(f, 0, SEEK_SET);

  vm.memory_size = END_OFFSET;
  if ((vm.memory = malloc(vm.memory_size)) == NULL) {
    fprintf(stderr, "Unable to allocate memory.\n");
    return 4;
  }

  memset(vm.memory, 0, vm.memory_size);

  int tmp = fread(&vm.memory[PROG_OFFSET], file_length, 1, f);
  if (tmp != 1) {
    fprintf(stderr, "Error reading file.\n");
    return 5;
  }

  if (strlen(argv[2]) > MEMORY_REGION_SIZE) {
    fprintf(stderr, "Userdata too large.\n");
    return 6;
  }

  strcpy(&vm.memory[USERDATA_OFFSET], argv[2]);

  memset(vm.regs, 0, sizeof(vm.regs));

  vm.regs[0] = USERDATA_OFFSET;
  vm.regs[1] = strlen(argv[2]);
  vm.regs[2] = RAM_OFFSET;
  vm.regs[3] = USERDATA_OFFSET - RAM_OFFSET;

  vm.regs[REG_IP] = PROG_OFFSET;
  vm.flags = 0;

  if (run_vm(&vm) != 0) {
    fprintf(stderr, "Error running VM.\n");
    return 7;
  }

  if (vm.regs[0] > vm.memory_size ||
      vm.regs[1] > vm.memory_size ||
      vm.regs[0] + vm.regs[1] > vm.memory_size ) {
    fprintf(stderr, "Invalid output memory region.\n");
    return 8;
  }

  for (i = 0; i < vm.regs[1]; i++) {
    printf("%02hhx", vm.memory[vm.regs[0] + i]);
  }
  putchar('\n');
}
